"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const CBOR = require("cbor");
class BundleBuilder {
    constructor(primaryURL) {
        this.primaryURL = primaryURL;
        this.sectionLengths = [];
        this.sections = [];
        this.responses = [];
        this.index = {};
        this.currentResponsesOffset = 0;
        validateExchangeURL(primaryURL);
    }
    createBundle() {
        if (!this.index[this.primaryURL]) {
            throw new Error(`Exchange for primary URL (${this.primaryURL}) does not exist`);
        }
        this.addSection("index", this.fixupIndex());
        this.addSection("responses", this.responses);
        let wbn = CBOR.encodeCanonical(this.createTopLevel());
        // Fill in the length field.
        let view = new DataView(wbn.buffer, wbn.byteOffset + wbn.length - 8);
        view.setUint32(0, Math.floor(wbn.length / 0x100000000));
        view.setUint32(4, wbn.length & 0xffffffff);
        return wbn;
    }
    addExchange(url, status, headers, payload) {
        validateExchangeURL(url);
        if (typeof payload === 'string')
            payload = byteString(payload);
        this.addIndexEntry(url, this.addResponse(new HeaderMap(status, headers), payload));
    }
    setManifestURL(url) {
        validateExchangeURL(url);
        this.addSection('manifest', url);
    }
    addSection(name, content) {
        if (this.sectionLengths.includes(name)) {
            throw new Error('Duplicated section: ' + name);
        }
        this.sectionLengths.push(name);
        this.sectionLengths.push(encodedLength(content));
        this.sections.push(content);
    }
    addResponse(headerMap, payload) {
        if (payload.length > 0 && !headerMap.has('content-type')) {
            throw new Error('Non-empty exchange must have Content-Type header');
        }
        let response = [new Uint8Array(CBOR.encodeCanonical(headerMap)), payload];
        this.responses.push(response);
        return encodedLength(response);
    }
    addIndexEntry(url, responseLength) {
        this.index[url] = [
            new Uint8Array(0),
            this.currentResponsesOffset,
            responseLength,
        ];
        this.currentResponsesOffset += responseLength;
    }
    fixupIndex() {
        // Adjust the offsets by the length of the response section's CBOR header.
        let responsesHeaderSize = encodedLength(this.responses.length);
        for (let key in this.index) {
            this.index[key][1] += responsesHeaderSize;
        }
        return this.index;
    }
    createTopLevel() {
        return [
            byteString('🌐📦'),
            byteString('b1\0\0'),
            this.primaryURL,
            new Uint8Array(CBOR.encodeCanonical(this.sectionLengths)),
            this.sections,
            new Uint8Array(8),
        ];
    }
}
exports.BundleBuilder = BundleBuilder;
;
class HeaderMap extends Map {
    constructor(status, headers) {
        super();
        if (status < 100 || status > 999) {
            throw new Error('Invalid status code');
        }
        this.set(':status', status.toString());
        for (let key in headers) {
            this.set(key.toLowerCase(), headers[key]);
        }
    }
    // This tells the CBOR library how to serialize this object.
    encodeCBOR(encoder) {
        // Convert keys and values to Uint8Array, as the CBOR representation of
        // header map is {bytestring => bytestring}.
        let m = new Map();
        for (let [key, value] of this.entries()) {
            m.set(byteString(key), byteString(value));
        }
        return encoder.pushAny(m);
    }
}
// Throws an error if `urlString` is not a valid exchange URL, i.e. it must:
// - be absolute,
// - have http: or https: protocol, and
// - have no credentials (user:password@) or hash (#fragment).
function validateExchangeURL(urlString) {
    let url = new URL(urlString);
    if (url.protocol !== 'http:' && url.protocol !== 'https:') {
        throw new Error("Exchange URL's protocol must be http(s): " + urlString);
    }
    if (url.username !== '' || url.password !== '') {
        throw new Error('Exchange URL must not have credentials: ' + urlString);
    }
    if (url.hash !== '') {
        throw new Error('Exchange URL must not have a hash: ' + urlString);
    }
}
// Returns the length of `value` when CBOR-encoded.
function encodedLength(value) {
    // We don't need to use canonical encoding here.
    return CBOR.encode(value).byteLength;
}
function byteString(s) {
    return Buffer.from(s, 'utf-8');
}
